package io.swagger.codegen.languages;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;

import io.swagger.codegen.CliOption;
import io.swagger.codegen.CodegenModel;
import io.swagger.codegen.CodegenOperation;
import io.swagger.codegen.CodegenParameter;
import io.swagger.codegen.CodegenProperty;
import io.swagger.codegen.DefaultCodegen;
import io.swagger.codegen.SupportingFile;
import io.swagger.codegen.mustache.CamelCaseLambda;
import io.swagger.models.Operation;
import io.swagger.models.Path;
import io.swagger.models.Swagger;
import io.swagger.models.properties.FileProperty;
import io.swagger.models.properties.ObjectProperty;
import io.swagger.models.properties.Property;

public class NetappOTSDTypescriptClient extends AbstractTypeScriptClientCodegen {
  public static final String TAGGED_UNIONS ="taggedUnions";
  private Map<String, Object> allModels;
  private static final String OVERRIDE_TAGS = "customOrganizeApis";
  private static final String API_FAMILY_KEY = "x-netapp-otsd-apiFamilyName";
  private static final String IS_OP_GET_KEY = "x-netapp-otsd-isOperationGet";
  private static final String IS_OP_LIST_KEY = "x-netapp-otsd-isOperationList";
  private static final String IS_OP_POST_KEY = "x-netapp-otsd-isOperationPost";
  private static final String IS_OP_PATCH_KEY = "x-netapp-otsd-isOperationPatch";
  private static final String IS_OP_PUT_KEY = "x-netapp-otsd-isOperationPut";
  private static final String IS_OP_DELETE_KEY = "x-netapp-otsd-isOperationDelete";
  private static final String CUSTOM_OP_NAME_KEY = "x-netapp-otsd-customMethodName";
  private static final String CUSTOM_RESOURCE_PATH_KEY = "x-netapp-otsd-customResourcePath";
  private static final String PARAM_ENUM_NAME_KEY = "x-netapp-otsd-paramEnumName";
  private static final String FIELDNAME_RECORDS = "records";
  private static final String FIELDNAME_RECORD = "record";
  private static final String FIELDNAME_JOB = "job";

  private boolean taggedUnions = false;

    public NetappOTSDTypescriptClient() {
        super();
        this.outputFolder = "generated-code/netapp-otsd-typescript";
        embeddedTemplateDir = templateDir = "netapp-otsd-typescript";
        modelTemplateFiles.put("model.mustache", ".ts");
        apiTemplateFiles.put("api.service.mustache", ".ts");
        apiPackage = "api";
        modelPackage = "model";

        typeMapping.put("file", "any");
        typeMapping.put("Blob", "any");

        this.cliOptions.add(CliOption.newBoolean(OVERRIDE_TAGS,
            "This option will organize API and corresponding operations according to the described family name from YAML")
            .defaultValue(Boolean.FALSE.toString()));
        additionalProperties.put("camelcase", new CamelCaseLambda());
    }

    @Override
    public String getName() {
        return "netapp-otsd-typescript";
    }

    @Override
    public String getHelp() {
        return "Generates Typescript client code for Angular 2.x - NetApp ONTAP Select Deploy.";
    }

    @Override
    public void processOpts() {
        super.processOpts();
        supportingFiles.add(new SupportingFile("models.mustache", modelPackage().replace('.', File.separatorChar), "models.ts"));
        supportingFiles.add(new SupportingFile("base.api.mustache", apiPackage().replace('.', File.separatorChar), "baseApi.Service.ts"));
        supportingFiles.add(new SupportingFile("deploy.api.service.module.mustache",
            apiPackage().replace('.', File.separatorChar), "deployApiService.module.ts"));
    }

    private String getIndexDirectory() {
        String indexPackage = modelPackage.substring(0, Math.max(0, modelPackage.lastIndexOf('.')));
        return indexPackage.replace('.', File.separatorChar);
    }

    @Override
    public boolean isDataTypeFile(final String dataType) {
        return dataType != null && dataType.equals("Blob");
    }

    @Override
    public String getTypeDeclaration(Property p) {
        if (p instanceof FileProperty) {
            return "Blob";
        } else if (p instanceof ObjectProperty) {
            return "any";
        } else {
            return super.getTypeDeclaration(p);
        }
    }


    @Override
    public String getSwaggerType(Property p) {
        String swaggerType = super.getSwaggerType(p);
        if (isLanguagePrimitive(swaggerType) || isLanguageGenericType(swaggerType)) {
            return swaggerType;
        }
        applyLocalTypeMapping(swaggerType);
        return swaggerType;
    }

    private String applyLocalTypeMapping(String type) {
        if (typeMapping.containsKey(type)) {
            type = typeMapping.get(type);
        }
        return type;
    }

    private boolean isLanguagePrimitive(String type) {
        return languageSpecificPrimitives.contains(type);
    }

    private boolean isLanguageGenericType(String type) {
        for (String genericType : languageGenericTypes) {
            if (type.startsWith(genericType + "<")) {
                return true;
            }
        }
        return false;
    }

    @Override
    public void postProcessParameter(CodegenParameter parameter) {
        super.postProcessParameter(parameter);
        parameter.dataType = applyLocalTypeMapping(parameter.dataType);
    }

    @Override
    public void preprocessSwagger(Swagger swagger) {
      String overrideTags = (String) additionalProperties.get(OVERRIDE_TAGS);
      if (overrideTags == null || overrideTags.isEmpty() || overrideTags.equalsIgnoreCase("false")) {
        return;
      }
      for (String pathName: swagger.getPaths().keySet()) {
        Path path = swagger.getPaths().get(pathName);
        String newTagName = (String) path.getVendorExtensions().get(NetappOTSDTypescriptClient.API_FAMILY_KEY);
        for (Operation op: path.getOperations()) {
          List<String> newTagNameList = new ArrayList<String>();
          newTagNameList.add(newTagName);
        if (newTagName != null && !newTagName.isEmpty()) {
          op.setTags(newTagNameList);
        }
        }
      }
      super.preprocessSwagger(swagger);
    }

    @Override
    public Map<String, Object> postProcessAllModels(Map<String, Object> objs) {
        Map<String, Object> result = super.postProcessAllModels(objs);

        for (Map.Entry<String, Object> entry : result.entrySet()) {
            Map<String, Object> inner = (Map<String, Object>) entry.getValue();
            List<Map<String, Object>> models = (List<Map<String, Object>>) inner.get("models");
            for (Map<String, Object> mo : models) {
                CodegenModel cm = (CodegenModel) mo.get("model");
                if (taggedUnions) {
                    mo.put(TAGGED_UNIONS, true);
                    if (cm.discriminator != null && cm.children != null) {
                        for (CodegenModel child : cm.children) {
                            cm.imports.add(child.classname);
                        }
                    }
                    if (cm.parent != null) {
                        cm.imports.remove(cm.parent);
                    }
                }
                // Add additional filename information for imports
                mo.put("tsImports", toTsImports(cm, cm.imports));
            }
        }
        this.allModels = objs;
        return result;
    }

    private CodegenModel getModelByName(String modelName) {
      for (Map.Entry<String, Object> entry : allModels.entrySet()) {
            Map<String, Object> inner = (Map<String, Object>) entry.getValue();
            List<Map<String, Object>> models = (List<Map<String, Object>>) inner.get("models");
            for (Map<String, Object> mo : models) {
                CodegenModel cm = (CodegenModel) mo.get("model");
                if (cm.classname.equals(modelName)) {
                  return cm;
                }
            }
      }
      return null;
    }

    private List<Map<String, String>> toTsImports(CodegenModel cm, Set<String> imports) {
        List<Map<String, String>> tsImports = new ArrayList<>();
        for (String im : imports) {
            if (!im.equals(cm.classname)) {
                HashMap<String, String> tsImport = new HashMap<>();
                tsImport.put("classname", im);
                tsImport.put("filename", toModelFilename(im));
                tsImports.add(tsImport);
            }
        }
        return tsImports;
    }

    @Override
    public Map<String, Object> postProcessOperations(Map<String, Object> operations) {
        @SuppressWarnings("unchecked")
    Map<String, Object> objs = (Map<String, Object>) operations.get("operations");
    HashSet<String> modelImports = new HashSet<>();
        @SuppressWarnings("unchecked")
    List<CodegenOperation> ops = (List<CodegenOperation>) objs.get("operation");
        for (CodegenOperation op : ops) {
          resetOpVendorExtensions(op);

          CodegenModel returnType = getModelByName(op.returnType);
          customprocessOperations(op);
            switch (op.httpMethod) {
                case "GET":
                    op.httpMethod = "RequestMethod.Get";
                    op.vendorExtensions.put(NetappOTSDTypescriptClient.IS_OP_GET_KEY, true);
                    String url = op.vendorExtensions.get(NetappOTSDTypescriptClient.CUSTOM_RESOURCE_PATH_KEY).toString();
                    if (!url.isEmpty() && !url.endsWith("}")) {
	                    op.vendorExtensions.put(NetappOTSDTypescriptClient.IS_OP_GET_KEY, false);
	                    op.vendorExtensions.put(NetappOTSDTypescriptClient.IS_OP_LIST_KEY, true);
	                    setCustomReturnType(op, returnType, NetappOTSDTypescriptClient.FIELDNAME_RECORDS);
	                    // work-around to make sure that we check for both record and records field.
	                    setCustomReturnType(op, returnType, NetappOTSDTypescriptClient.FIELDNAME_RECORD);
                        break;
                  }
                  setCustomReturnType(op, returnType, NetappOTSDTypescriptClient.FIELDNAME_RECORD);
                    break;
                case "POST":
                    op.httpMethod = "RequestMethod.Post";
                    op.vendorExtensions.put(NetappOTSDTypescriptClient.IS_OP_POST_KEY, true);
                    setCustomReturnType(op, returnType, NetappOTSDTypescriptClient.FIELDNAME_JOB);
                    break;
                case "PUT":
                    op.httpMethod = "RequestMethod.Put";
                    op.vendorExtensions.put(NetappOTSDTypescriptClient.IS_OP_PUT_KEY, true);
                    setCustomReturnType(op, returnType, NetappOTSDTypescriptClient.FIELDNAME_JOB);
                    break;
                case "DELETE":
                    op.httpMethod = "RequestMethod.Delete";
                  op.vendorExtensions.put(NetappOTSDTypescriptClient.IS_OP_DELETE_KEY, true);
                  setCustomReturnType(op, returnType, NetappOTSDTypescriptClient.FIELDNAME_JOB);
                    break;
                case "PATCH":
                    op.httpMethod = "RequestMethod.Patch";
                  op.vendorExtensions.put(NetappOTSDTypescriptClient.IS_OP_PATCH_KEY, true);
                  setCustomReturnType(op, returnType, NetappOTSDTypescriptClient.FIELDNAME_JOB);
                    break;
                default:
                    throw new RuntimeException("Unknown HTTP Method " + op.httpMethod + " not allowed");
                }

          gatherModelImports(op, modelImports);
        }
        objs.put("modelImports", modelImports);
        return operations;
    }

    private void resetOpVendorExtensions(CodegenOperation op) {
      String[] operations = {NetappOTSDTypescriptClient.IS_OP_GET_KEY, NetappOTSDTypescriptClient.IS_OP_LIST_KEY,
          NetappOTSDTypescriptClient.IS_OP_POST_KEY, NetappOTSDTypescriptClient.IS_OP_PATCH_KEY,
          NetappOTSDTypescriptClient.IS_OP_PUT_KEY, NetappOTSDTypescriptClient.IS_OP_DELETE_KEY};
      for (String opName: operations) {
        op.vendorExtensions.put(opName, false);
      }
    }

    private void customprocessOperations(CodegenOperation operation) {
      addMethodName(operation);
      addResourcePath(operation);
    }

    private void addMethodName(CodegenOperation op) {
      /* It seems that operation ID is bound to the packages related to the API and bound to change and
      not a good candidate to derive method name. Keeping this for now. */
      String methodName = op.operationIdOriginal;
      String[] originalOperationIdArray = op.operationIdOriginal.split("\\.");
      if (originalOperationIdArray.length == 3) {
        String[] action = originalOperationIdArray[2].split("\\_");
        StringBuilder modifiedMethodName = new StringBuilder();
        for (String act : action) {
          modifiedMethodName.append(act + ".");
        }
        methodName = modifiedMethodName.toString();
      }
      else if (originalOperationIdArray.length > 3) {
	// workaround for capacity pool APIs. pls remove if YAML changes to decide
	String[] action = originalOperationIdArray[3].split("\\_");
	StringBuilder modifiedMethodName = new StringBuilder();
	for (String act : action) {
		modifiedMethodName.append(act + ".");
	}
	methodName = modifiedMethodName.toString();
      }
      op.vendorExtensions.put(NetappOTSDTypescriptClient.CUSTOM_OP_NAME_KEY, DefaultCodegen.camelize(methodName, true));
    }

    private void addResourcePath(CodegenOperation op) {
      String tempPath = op.path;
      tempPath = tempPath.replaceAll("\\{", ":\\{").replaceAll(":", Matcher.quoteReplacement("$"));
      op.vendorExtensions.put(NetappOTSDTypescriptClient.CUSTOM_RESOURCE_PATH_KEY, tempPath.toString());
    }

    private void setCustomReturnType(CodegenOperation op, CodegenModel opReturnType, String fieldName) {
      if (opReturnType ==  null) {
        return;
      }
      for (CodegenProperty cp: opReturnType.allVars) {
        if (!cp.baseName.equalsIgnoreCase(fieldName)) {
          continue;
        }
        op.returnBaseType = cp.datatype;
        op.returnSimpleType = true;

        if (!cp.datatype.contains("Array")) {
          return;
        }
        op.returnSimpleType = false;
        String[] returnTypeChunks = op.returnBaseType.split("\\<");
        if (returnTypeChunks.length < 1) {
          return;
        }
        returnTypeChunks = returnTypeChunks[1].split("\\>");
        if (returnTypeChunks.length > 0) {
          op.returnBaseType = returnTypeChunks[0];
          return;
        }
      }
    }

    private void gatherModelImports(CodegenOperation op, HashSet<String> modelImports) {
      for (CodegenParameter param : op.allParams) {
        if (param.isEnum) {
            param.vendorExtensions.put(NetappOTSDTypescriptClient.PARAM_ENUM_NAME_KEY,
                param.datatypeWithEnum + op.vendorExtensions.get(NetappOTSDTypescriptClient.CUSTOM_OP_NAME_KEY));
          }
          if (!param.isPrimitiveType && !param.isContainer && !param.dataType.equals("any") && !param.dataType.equals("Blob")) {
            modelImports.add(param.dataType);
          }
      }

      if (op.returnType != null && !op.returnBaseType.equals(op.returnType) && !op.returnTypeIsPrimitive) {
        modelImports.add(op.returnType);
      }

      if (op.returnBaseType != null && !op.returnTypeIsPrimitive) {
        modelImports.add(op.returnBaseType);
      }
    }

    @Override
    public Map<String, Object> postProcessModels(Map<String, Object> objs) {
        Map<String, Object> result = super.postProcessModels(objs);

        return postProcessModelsEnum(result);
    }

    @Override
    public String toApiName(String name) {
        if (name.length() == 0) {
            return "DefaultService";
        }
        return initialCaps(name) + "Service";
    }

    @Override
    public String toApiFilename(String name) {
        if (name.length() == 0) {
            return "default.service";
        }
        return camelize(name, true) + ".service";
    }

    @Override
    public String toApiImport(String name) {
        return apiPackage() + "/" + toApiFilename(name);
    }

    @Override
    public String toModelFilename(String name) {
        return camelize(toModelName(name), true);
    }

    @Override
    public String toModelImport(String name) {
        return modelPackage() + "/" + toModelFilename(name);
    }

    private String getApiFilenameFromClassname(String classname) {
        String name = classname.substring(0, classname.length() - "Service".length());
        return toApiFilename(name);
    }

}
